//*****************************************************************************
//*****************************************************************************
//  FILENAME:   sensor.c
//  Version: 2.0, 2012/10/02
//  Generated by Urs Bucher
//
//  DESCRIPTION:  sensor specific functions 
//
//-----------------------------------------------------------------------------
// Originally authored by Toradex AG. This work is now available under the terms and conditions
// of the Creative Commons License 'Attribution CC BY', details of which can be found here:
// http://creativecommons.org/licenses/by/3.0/
//*****************************************************************************
//*****************************************************************************
//#define __DEBUG__
#include <m8c.h>        // part specific constants and macros
#include "sensor.h"
#include "psocgpioint.h"
#include "psocapi.h"
#include "feature.h"
#include "parameters.h"
#include "types_n_macros.h"
//----------------------------------------
// Sensor parameters
// Factory Defaults are defined here (NUM_xxx are defined in "sensor.h")
//----------------------------------------
const BYTE FactoryDefault_Param_8   [NUM_PARAM_8] =
{
	0x00,			// report mode
	0x00,			// LED mode
	GP2Y0A60SZ0F	// Sensor Type is GP2Y0A60SZ0F (80cm)
};

const WORD FactoryDefault_Param_16  [NUM_PARAM_16] = 
{
	500,		// report rate [ms]
	500,		// sample rate [ms]
	50000,		// sample rate [us]	(not used for this sensor)
	0,			// Offset
	0,			// User data 1
	0			// User data 2
};

const DWORD FactoryDefault_Param_32  [NUM_PARAM_32] =
{
	0x00000000	// currently no 32bit parameters
};

const tParam_168 FactoryDefault_Param_168 [NUM_PARAM_168] =
{
	"Distance Sensor",	// user device name
	"Frame Number",		// user name Channel 0
	"Distance"			// user name Channel 1
};


//----------------------------------------
// this module's global variables
//----------------------------------------
tReport Report;

//----------------------------------------
// constant lookup-table for ADC->physical value conversion
//----------------------------------------

//----------------------------------------
// type definitions
//----------------------------------------

typedef enum 
{	
	eReset,
	eIdle,
	eReadADC,
	eCalculating
} tFSM_Measure;

//----------------------------------------
// variables

tFSM_Measure	FSM;
WORD			Countdown_1ms;
WORD			Raw_Offset; 

BOOL			NewSample = FALSE;

// store original drivemodes during suspend
BYTE			DM0_2, DM0_1, DM0_0;
BYTE			DM2_2, DM2_1, DM2_0;
BYTE			ARF;

//----------------------------------------
// function prototypes
static int CalculateDistance (WORD ADC);

//-----------------------------------------------------------------------------
// Initializes the Sensor 
//-----------------------------------------------------------------------------
void Sensor_Start ()
{
	// turn on power to the sensor
	GPIO_LOW(OFF_nON)
	
	// start ADC sampling
	PGA_Start (PGA_HIGHPOWER);
	DelSig_Start (DelSig_MEDPOWER);
	DelSig_StartAD(); 		// Start continous sampling
	
	FSM = eReset;
}
//-----------------------------------------------------------------------------
// Prepare Sensor for Suspend
//-----------------------------------------------------------------------------
void Sensor_Suspend()
{
	// stop and power off ADC
	DelSig_StopAD();
	DelSig_Stop();
	PGA_Stop();
	// turn off power for the sensor
	GPIO_HIGH(OFF_nON)

	// store original drivemodes during  suspend
	DM0_2 = PRT0DM2;
	DM0_1 = PRT0DM1;
	DM0_0 = PRT0DM0;
	DM2_2 = PRT2DM2;
	DM2_1 = PRT2DM1;
	DM2_0 = PRT2DM0;
	// store Analog power settings
	ARF   = ARF_CR;
	
	// set all I/Os to/from the sensor to High-Z Analog (=110), 
	// except the OFF_nON signal. 
	// OFF_nON is on Port0, all other signals on Port2
	PRT0DM2  = 0xFF;
	PRT0DM1  = 0xFF;
	PRT0DM0  = 0x00;
	PRT2DM2 |= ~OFF_nON_MASK;
	PRT2DM1 |= ~OFF_nON_MASK;
	PRT2DM0 &=  OFF_nON_MASK;

}
//-----------------------------------------------------------------------------
// Resume Sensor from Suspend mode
//-----------------------------------------------------------------------------
void Sensor_Resume()
{
	BYTE ReadMode;
	
	// restore drivemodes for GPIOs to/from the sensor
	PRT2DM2 = DM2_2;
	PRT2DM1 = DM2_1;
	PRT2DM0 = DM2_0;

	PRT0DM2 = DM0_2;
	PRT0DM1 = DM0_1;
	PRT0DM0 = DM0_0;
	
	ARF_CR = ARF;

	// turn on power for the sensor
	Sensor_Start();
	FSM = eReset;
}


//-----------------------------------------------------------------------------
// Function that will be called when a FEATURE REPORT is received that 
// cannot be handled with the standard functionality in feature.c
// The feature report data is available in the "FEATURE_RPT" variable
//
// --> see feature.c for an example how to parse FEATURE_RPT
//
//	return Value: number of Parameter bytes (1-31) that are 
//				  returned in the IN report
//-----------------------------------------------------------------------------
BYTE Sensor_Specific_Feature(void)
{
	// currently no special feature reports are handled
	return 0;
					}

//-----------------------------------------------------------------------------
// Function that will be called when a FEATURE REPORT is received that 
// sets / gets a 8/16/32/168bit parameter in feature.c
// The feature report data is available in the "FEATURE_RPT" variable
//
// --> see feature.c for an example how to parse FEATURE_RPT
//-----------------------------------------------------------------------------
void Sensor_Specific_Parameter(BYTE *NumberOfReturnedBytes)
{
	// currently no specific parameters are handled
}

//-----------------------------------------------------------------------------
// Read ADC 
//-----------------------------------------------------------------------------
static WORD GP2D12_ReadADC (void)
{
	WORD Data;

	if ( DelSig_fIsDataAvailable() )
	{
		M8C_DisableGInt;
		Data = DelSig_wGetDataClearFlag();	
		M8C_EnableGInt;
		
		return Data;
	}
	else
		return 0xFFFF;
}

//-----------------------------------------------------------------------------
// void Sensor_FSM_Measure(void)
//
//	finite state machine to control the measurement process for 
//  the Sensor
//
//  IMPORTANT: 	call this routine once every millisecond. Otherwise the timing
//				parameters cannot be met.
//-----------------------------------------------------------------------------
void Sensor_FSM_Measure() 
{
	static WORD ADC; // ADC value
	WORD 		Distance;
	BYTE		SensorType;
	
	WORD ADCdbg;
	
	ADCdbg = ADC;
	// decrement millisecond counter
	if (Countdown_1ms != 0)
		Countdown_1ms--;

	// process actual state machine
	switch (FSM)
	{
		case eReset:
			Countdown_1ms = 10; // 10ms delay until first measurement finished
			PGA_SetGain(PGA_G0_56);
			FSM = eIdle;
			break;
			
		case eIdle: 
			if (Countdown_1ms == 0)
			{
				Get_Param_16(PAR16_SAMPLERATE_MS, VOLATILE, &Countdown_1ms);
				FSM = eReadADC;					  
			}
			break;			
			
		case eReadADC:
			ADC = GP2D12_ReadADC();
			FSM = eCalculating;

			// Fill-in raw value
			break;			

		case eCalculating:
			{
				Get_Param_8(PAR8_SENSORTYPE, VOLATILE, &SensorType);
	
				switch(SensorType)
				{
					case GP2Y0A60SZ0F:
						{
							Distance = CalculateDistance (ADC);
			
							Report.Ch1_LSB = (BYTE)(Distance & 0xFF);
							Report.Ch1_MSB = (BYTE)(Distance / 0x100);
						}
						break;									

					case RAWVALUE:
						{
							Report.Ch1_LSB = (BYTE)(ADC & 0xFF);
							Report.Ch1_MSB = (BYTE)(ADC / 0x100);
						}
						break;

					default:
						break;
				}
				NewSample = TRUE;
				FSM = eIdle;
			}
			break;			
	}
}

//-----------------------------------------------------------------------------
// Routine: int CalculateDistance (WORD ADC)
//
// Description: Calculate the distance from the ADC output of a IR distance sensor
//
// Return: WORD - distance [0.1mm]
//
//-----------------------------------------------------------------------------
extern const WORD ADC2L[TABLESIZE];	// for GP2Y0A60SZ0F

static int CalculateDistance (WORD ADC)
{
	BYTE SensorType;
	WORD index;
	WORD L_Below, L_Above, sum = 0;
	INT  Temp;
	
	Get_Param_16(PAR16_OFFSET, VOLATILE, &Raw_Offset);

	Temp = (INT)(ADC&0x7FFF);			// make sur that ADC is positiv
	Temp = Temp + (INT)Raw_Offset;		// add offset (offset is intepreted as signed)

	ADC = (WORD)Temp;					// convert ist back to unsisigned 			
	
	index = ADC >> 3;					// use every 8th value for the look up

	// all values below 288 give the minumum distance 
	if (index<=TABLESTA)				// ADC2L starts with values for 288 (8x36) and more
	{
		index = 0;
	}		
	else
	{
		index = index-TABLESTA;	
		// all values above 3416 give the maximum distance 
		if (index>(TABLESIZE-1-1))		// index > then array -1
		{
			index = TABLESIZE-1-1;		
		}
	}		
			
	L_Below = ADC2L[index];
	L_Above = ADC2L[index+1];
	
	// do linear interpolation, assuming every 8th data is in the lookup table
	sum = L_Below;								// sum ~ L
	sum += (ADC & 0x0001) ? L_Above : L_Below;  // sum ~ 2xL

	L_Above <<=1; L_Below <<=1;
	sum += (ADC & 0x0002) ? L_Above : L_Below;	// sum ~ 4xL
	sum >>=1;									// sum ~ 2xL
	sum += (ADC & 0x0004) ? L_Above : L_Below;	// sum ~ 4xL
	
	return sum >>2;
}

//-----------------------------------------------------------------------------
const WORD ADC2L[] =	// table lists every 8th distance value
{
	12167,			// ADC=240  Index = ADC/8-TABLESTA = 0
	11722,
	11278,
	10833,
	10389,
	9944,
	9500,
	9056,
	8638,
	8187,
	7600,
	7000,
	6666,
	6383,
	6158,
	6000,
	5884,
	5778,
	5681,
	5592,
	5510,
	5433,
	5361,
	5293,
	5227,
	5162,
	5098,
	5033,
	4967,
	4901,
	4835,
	4770,
	4706,
	4643,
	4581,
	4520,
	4461,
	4403,
	4346,
	4291,
	4238,
	4186,
	4136,
	4089,
	4043,
	4000,
	3958,
	3917,
	3876,
	3836,
	3797,
	3758,
	3720,
	3682,
	3645,
	3608,
	3572,
	3537,
	3502,
	3468,
	3434,
	3401,
	3369,
	3337,
	3306,
	3275,
	3245,
	3215,
	3186,
	3158,
	3130,
	3103,
	3076,
	3050,
	3025,
	3000,
	2976,
	2952,
	2928,
	2906,
	2883,
	2861,
	2840,
	2819,
	2798,
	2778,
	2758,
	2738,
	2719,
	2700,
	2681,
	2663,
	2645,
	2627,
	2610,
	2593,
	2576,
	2559,
	2543,
	2526,
	2510,
	2494,
	2478,
	2462,
	2446,
	2431,
	2415,
	2400,
	2384,
	2369,
	2354,
	2339,
	2324,
	2309,
	2294,
	2280,
	2265,
	2251,
	2237,
	2223,
	2209,
	2196,
	2182,
	2169,
	2156,
	2143,
	2130,
	2118,
	2105,
	2093,
	2081,
	2070,
	2058,
	2047,
	2036,
	2026,
	2015,
	2005,
	1995,
	1985,
	1975,
	1966,
	1956,
	1946,
	1937,
	1927,
	1918,
	1908,
	1899,
	1890,
	1880,
	1871,
	1862,
	1853,
	1844,
	1835,
	1826,
	1817,
	1809,
	1800,
	1791,
	1783,
	1774,
	1766,
	1758,
	1749,
	1741,
	1733,
	1725,
	1717,
	1709,
	1701,
	1694,
	1686,
	1678,
	1671,
	1663,
	1656,
	1648,
	1641,
	1634,
	1627,
	1620,
	1613,
	1606,
	1599,
	1593,
	1586,
	1579,
	1573,
	1566,
	1560,
	1554,
	1548,
	1542,
	1536,
	1530,
	1524,
	1518,
	1512,
	1507,
	1501,
	1496,
	1491,
	1485,
	1480,
	1475,
	1469,
	1464,
	1459,
	1454,
	1449,
	1443,
	1438,
	1433,
	1428,
	1423,
	1418,
	1413,
	1408,
	1403,
	1399,
	1394,
	1389,
	1384,
	1379,
	1375,
	1370,
	1365,
	1361,
	1356,
	1351,
	1347,
	1342,
	1338,
	1333,
	1329,
	1324,
	1320,
	1315,
	1311,
	1306,
	1302,
	1298,
	1293,
	1289,
	1285,
	1281,
	1276,
	1272,
	1268,
	1264,
	1260,
	1255,
	1251,
	1247,
	1243,
	1239,
	1235,
	1231,
	1227,
	1223,
	1219,
	1215,
	1211,
	1207,
	1203,
	1199,
	1195,
	1191,
	1188,
	1184,
	1180,
	1176,
	1172,
	1168,
	1165,
	1161,
	1157,
	1153,
	1149,
	1146,
	1142,
	1138,
	1135,
	1131,
	1127,
	1124,
	1120,
	1116,
	1113,
	1109,
	1105,
	1102,
	1098,
	1094,
	1091,
	1087,
	1084,
	1080,
	1076,
	1073,
	1069,
	1066,
	1062,
	1059,
	1055,
	1052,
	1048,
	1044,
	1041,
	1037,
	1034,
	1030,
	1027,
	1023,
	1020,
	1016,
	1013,
	1009,
	1006,
	1002,
	998,
	994,
	990,
	986,
	982,
	978,
	974,
	969,
	965,
	961,
	957,
	953,
	949,
	944,
	940,
	936,
	932,
	928,
	924,
	920,
	915,
	911,
	907,
	903,
	899,
	895,
	890,
	886,
	882,
	878,
	874,
	870,
	866,
	861,
	857,
	853,
	849,
	845,
	841,
	836,
	832,
	828,
	824,
	820,
	816,
	812,
	807,
	803,
	799,
	795,
	791,
	787,
	782,
	778,
	774,
	770,
	766,
	762,
	758,
	753,
	749,
	745,
	741,
	737,
	733,
	728,
	724,
	720,
	716,
	712,
	708,
	704,
	699,
	695,
	691,
	687,
	687,
	687
};
	

#if 0  // old Tabelle Mechantronics
//-----------------------------------------------------------------------------
const WORD ADC2L[] =	// table lists every 8th distance value
{
10593,  // ADC 600 (75x8) 1.0593m
10275,  // ADC 608 (76x8) 1.0275m
9974,
9688,
9417,
9158,
8913,
8678,
8455,
8242,
8038,
7844,
7658,
7479,
7309,
7145,
6988,
6837,
6693,
6553,
6420,
6291,
6167,
6047,
5932,
5821,
5714,
5611,
5511,
5414,
5321,
5231,
5143,
5059,
4977,
4898,
4821,
4746,
4674,
4604,
4535,
4469,
4405,
4342,
4282,
4223,
4165,
4109,
4055,
4002,
3950,
3900,
3851,
3803,
3757,
3711,
3667,
3624,
3581,
3540,
3500,
3461,
3422,
3385,
3348,
3312,
3277,
3243,
3209,
3176,
3144,
3113,
3082,
3052,
3022,
2993,
2965,
2937,
2910,
2883,
2857,
2831,
2806,
2781,
2757,
2733,
2709,
2686,
2664,
2642,
2620,
2599,
2578,
2557,
2537,
2517,
2497,
2478,
2459,
2440,
2422,
2404,
2386,
2369,
2352,
2335,
2318,
2302,
2286,
2270,
2254,
2239,
2224,
2209,
2194,
2180,
2165,
2151,
2137,
2124,
2110,
2097,
2084,
2071,
2058,
2046,
2033,
2021,
2009,
1997,
1985,
1974,
1962,
1951,
1940,
1929,
1918,
1907,
1897,
1886,
1876,
1866,
1856,
1846,
1836,
1826,
1817,
1807,
1798,
1789,
1780,
1771,
1762,
1753,
1744,
1736,
1727,
1719,
1710,
1702,
1694,
1686,
1678,
1670,
1662,
1655,
1647,
1640,
1632,
1625,
1617,
1610,
1603,
1596,
1589,
1582,
1575,
1568,
1562,
1555,
1549,
1542,
1536,
1529,
1523,
1517,
1510,
1504,
1498,
1492,
1486,
1480,
1474,
1469,
1463,
1457,
1452,
1446,
1440,
1435,
1429,
1424,
1419,
1413,
1408,
1403,
1398,
1393,
1388,
1383,
1378,
1373,
1368,
1363,
1358,
1353,
1349,
1344,
1339,
1335,
1330,
1326,
1321,
1317,
1312,
1308,
1304,
1299,
1295,
1291,
1287,
1282,
1278,
1274,
1270,
1266,
1262,
1258,
1254,
1250,
1246,
1242,
1239,
1235,
1231,
1227,
1224,
1220,
1216,
1213,
1209,
1205,
1202,
1198,
1195,
1191,
1188,
1184,
1181,
1178,
1174,
1171,
1168,
1164,
1161,
1158,
1154,
1151,
1148,
1145,
1142,
1139,
1136,
1133,
1129,
1126,
1123,
1120,
1117,
1115,
1112,
1109,
1106,
1103,
1100,
1097,
1094,
1092,
1089,
1086,
1083,
1081,
1078,
1075,
1072,
1070,
1067,
1065,
1062,
1059,
1057,
1054,
1052,
1049,
1047,
1044,
1042,
1039,
1037,
1034,
1032,
1029,
1027,
1024,
1022,
1020,
1017,
1015,
1013,
1010,
1008,
1006,
1004,
1001,
999,
997,
995,
992,
990,
988,
986,
984,
982,
979,
977,
975,
973,
971,
969,
967,
965,
963,
961,
959,
957,
955,
953,
951,
949,
947,
945,
943,
941,
939,
937,
935,
933,
932,
930,
928,
926,
924,
922,
921,
919,
917,
915,
913,
912,
910,
908,
906,
905,
903,
901,
899,
898,
896,
894,
893,
891,
889,
888,
886,
884,
883,
881,
879,
878,
876,
875,
873,
871,
870,
868,
867,
865,
864,
862,
861,
859,
858,
856,
855,
853,
852,
850,
849,
847,
846,
844,
843,
841,
840,
838,
837,
836,
834,
833,
831,
830,
829,
827,
826,
824,
823,
822,
820,
819,
818,
816,
815,
814,
812,
811,
810,
808
};
#endif

//-----------------------------------------------------------------------------
void Sensor_Debug(void)
{
/* currently no debug function */
}