/////////////////////////////////////////////////////////////////////////////
// Oak.cpp
//
// Control for Toradex Oak USB Sensors
//
// Written by Andy Kiser <andy.kiser@toradex.com>
// Copyright (c) 2007. All Rights Reserved.
//
// This file is provided "as is" with no expressed or implied warranty.
// The author accepts no liability for any damage/loss of business that
// this product may cause.
//
//////////////////////////////////////////////////////////////////////

#pragma once
#include <afxwin.h>		// diagnostic support
#include <windows.h>	// Master include file for Windows applications.
#include <tchar.h>
#include <vector>
#include <string.h>
#include "oak.h"

#include <dbt.h>		// quates for WM_DEVICECHANGE and BroadcastSystemMessage

#ifdef __cplusplus	// If used by C++ code, 
	extern "C"			// we need to export the C interface
	{          
#endif
	#include "hidsdi.h"
	#include <setupapi.h>
#ifdef __cplusplus	
	}
#endif

// common precompiler macros
#define ARRAYLEN(ary) ( sizeof(ary)/sizeof(ary[0]) )

// module specific types
typedef enum {eDeviceName, eChannelName} tUserStrType;

typedef struct
{
	TCHAR		DevicePath[255];

	WORD		VID;
	WORD		PID;
	WORD		REV;
	TCHAR		SN[21];

	TCHAR		DeviceName[21];
	TCHAR		UserDeviceName[21];
	TCHAR		ChannelName[21];
	TCHAR		UserChannelName[21];

	tNotifyFct	OnConnect;
	tNotifyFct	OnDisconnect;
	LPARAM		UserParam;	// this parameter will be passed back to the loop back function
}	tOakFilter;


// function prototypes
DWORD WINAPI Oak_HotplugHandler( DWORD dwControl, DWORD dwEventType, LPVOID lpEventData, LPVOID lpContext );
static void SensorList_Add (tOakSensor *NewSensor);
static void SensorList_Destroy();
static void SensorList_RemoveSensor(TCHAR *RemovePath);
static tOakSensor *SensorList_FindPath(TCHAR *ComparePath);
static void __DEBUG(void);  // The function is implemented at the end of this file

// static variables 
std::vector<tOakFilter>	Filter_OnDevChange;	// holds filters that are registered to generate connect/disconnect events
//-------------------------------------------------------------------------------------------------------------------------
static void Oak_DestroyChannel(tChannel *Channel)
// free all memory allocated for a channel
// reset all pointers to NULL
{	
//nothing to do
}

//-------------------------------------------------------------------------------------------------------------------------
// Get a User Device Name or User Channel Name from the Oak Sensor.
// return value: String length (Nr of characters without trailing zero)
#define VOLATILE TRUE
#define NONVOLATILE FALSE
static size_t Oak_Retreive_UserString(LPTSTR DevicePath, char Index, bool Volatile, TCHAR *buffer, BYTE MaxLen)
{
	BYTE	RptBuf[33];
	size_t	StrLen;

	// ---- Prepare OUT feature report to hold the appropriate command
	memset( RptBuf, 0x00, sizeof(RptBuf) );
	RptBuf[1]= 0x01;		// Get
	RptBuf[2]= Volatile? 0x00:0x01;  
	RptBuf[3]= 21;			// Lnge
	RptBuf[4]= Index;		// Index LSB (MSB = 0)

	// ---- Read Information from Sensor
	Oak_Feature(DevicePath, RptBuf, TRUE);

	// ---- copy String into output buffer, transform ASCII/Unicode if necessary;
	StrLen = strlen((char*)RptBuf+2);
	for (size_t i=0; (i<=StrLen)&&(i<MaxLen); i++)
		buffer[i] = RptBuf[i+2];

	return StrLen;
}
//-------------------------------------------------------------------------------------------------------------------------
void Oak_DestroySensor(OAK_HANDLE Sensor)
{	
// nothing to do
}

//-------------------------------------------------------------------------------------------------------------------------
static void Oak_Retreive_SensorInfo(LPTSTR DevicePath, OAK_HANDLE pSensor)
// retreives all information for a Sensor.
// All standard parameters (including pSensor->DevicePath) are filled in by this function.
{
	bool			bResult;
	WCHAR			wchBuffer[128];
	HANDLE			DeviceHandle;
	HIDD_ATTRIBUTES	Attributes;

	// ---- copy DevicePath, convert to lowercase for unambigous equality comparison 
	_tcsncpy_s(pSensor->DevicePath, ARRAYLEN(pSensor->DevicePath), DevicePath, _TRUNCATE);
	_tcslwr_s (pSensor->DevicePath, ARRAYLEN(pSensor->DevicePath));

	// Open a handle to the device.
	DeviceHandle = CreateFile(pSensor->DevicePath, 0x00, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);

	// ---- Get VID, PID and REV
	Attributes.Size = sizeof(Attributes);							
	bResult = HidD_GetAttributes(DeviceHandle, &Attributes) != 0;

	pSensor->VID = Attributes.VendorID;
	pSensor->PID = Attributes.ProductID;
	pSensor->REV = Attributes.VersionNumber;
	
	// ---- Get DeviceName
	if ( HidD_GetProductString(DeviceHandle, wchBuffer, sizeof(wchBuffer)) )
	{
		// copy Device name, convert UNICODE -> ASCII if necessary
		for (int i=0; i<ARRAYLEN(pSensor->DeviceName); i++)
			pSensor->DeviceName[i] = (TCHAR)wchBuffer[i];
		wchBuffer[ ARRAYLEN(wchBuffer)-1 ] = 0;
	}	
		//_tcsncpy_s(pSensor->DeviceName, sizeof(pSensor->DeviceName)/sizeof(TCHAR), wchBuffer, _TRUNCATE);
	
	// ---- Get Serial Number
	if ( HidD_GetSerialNumberString(DeviceHandle, wchBuffer, sizeof(wchBuffer)) )
	{
		// copy Serial number, convert UNICODE -> ASCII if necessary
		for (int i=0; i<ARRAYLEN(pSensor->SN); i++)
			pSensor->SN[i] = (TCHAR)wchBuffer[i];
			pSensor->SN[ ARRAYLEN(pSensor->SN)-1 ] = 0;
		//_tcsncpy_s(pSensor->SN, sizeof(pSensor->SN)/sizeof(TCHAR), wchBuffer, _TRUNCATE);
	
	}
	CloseHandle(DeviceHandle);
}


//-------------------------------------------------------------------------------------------------------------------------
static void Oak_Retreive_SpecificSensorInfo(OAK_HANDLE pSensor)
// retreives USB information specific to Oak Sensor devices
// When entering this function, the field pSensor->DevicePath needs to contain the sensor's DevicePath
// All other parameters are filled in by this function.
{
	HANDLE					DeviceHandle;
	PHIDP_PREPARSED_DATA	PreparsedData;
	HIDP_VALUE_CAPS			ValueCapabilities[MAX_NO_CHANNELS];
	WCHAR					wchBuffer[128];
	WCHAR					*wchToken, *wchNextToken;
	size_t					StrLen;
	TCHAR					Name[20];


	// ---- Open a handle to the device.
	DeviceHandle = CreateFile(pSensor->DevicePath, 0x00, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);

	// ---- Get User Device Name
	StrLen = Oak_Retreive_UserString(pSensor->DevicePath, 0, VOLATILE, Name, 20) +1;
	_tcsncpy_s(pSensor->UserDeviceName, ARRAYLEN(pSensor->UserDeviceName), Name, _TRUNCATE);
	StrLen = Oak_Retreive_UserString(pSensor->DevicePath, 0, NONVOLATILE, Name, 20) +1;
	_tcsncpy_s(pSensor->UserDeviceName_NV, ARRAYLEN(pSensor->UserDeviceName_NV), Name, _TRUNCATE);

	// ---- Get information about all channels
	HidD_GetPreparsedData(DeviceHandle, &PreparsedData);
	pSensor->NumChannels = MAX_NO_CHANNELS;
	if (HidP_GetValueCaps(HidP_Input, ValueCapabilities, &pSensor->NumChannels, PreparsedData) == HIDP_STATUS_SUCCESS)
	{
		for (int i=0; i<pSensor->NumChannels; i++)
		{
			// Get BitSize, Unit, Unit Exponent
			pSensor->Channel[i].BitSize = (BYTE)ValueCapabilities[i].BitSize;
			pSensor->Channel[i].Unit	= ValueCapabilities[i].Units;
			pSensor->Channel[i].UnitExponent = (char)ValueCapabilities[i].UnitsExp;
			if (pSensor->Channel[i].UnitExponent >=8)		// correction for negative exponents
					pSensor->Channel[i].UnitExponent -= 16;
			pSensor->Channel[i].IsSigned = (ValueCapabilities[i].LogicalMin < 0);

			// Get Channel Name...
			HidD_GetIndexedString(DeviceHandle, ValueCapabilities[i].NotRange.StringIndex, wchBuffer, sizeof(wchBuffer));
			wchToken = wcstok_s(wchBuffer, L"[]", &wchNextToken);	// clips wchBuffer to contain the first wchToken
			StrLen = wcslen(wchToken)+1; 
			while (wchBuffer[StrLen-2]==L' ')	// strip ending spaces
					StrLen--;
			if (StrLen > 1)
			{
				// copy Channel name, convert UNICODE -> ASCII if necessary
				for (int j=0; j< ARRAYLEN(pSensor->Channel[i].ChannelName); j++)
					pSensor->Channel[i].ChannelName[j] = (TCHAR)wchToken[j];
				// strip ending space
				StrLen = wcslen(wchToken);
				if (wchToken[StrLen-1] == L' ' )
					pSensor->Channel[i].ChannelName[StrLen-1] = 0;
				pSensor->Channel[i].ChannelName[ARRAYLEN(pSensor->Channel[i].ChannelName)-1] = 0;
				//	_tcsncpy_s(pSensor->Channel[i].ChannelName, ARRAYLEN(pSensor->Channel[i].ChannelName), wchToken, _TRUNCATE);
	
			//...  and Unit String
				wchToken = wcstok_s(NULL, L"[]", &wchNextToken);
				if (wchToken != NULL)
				{
					// copy unit string, convert UNICODE -> ASCII if necessary
					for (int j=0; j< ARRAYLEN(pSensor->Channel[i].UnitStr); j++)
						pSensor->Channel[i].UnitStr[j] = (TCHAR)wchToken[j];
					pSensor->Channel[i].UnitStr[ARRAYLEN(pSensor->Channel[i].UnitStr)-1] = 0;
					//_tcsncpy_s(pSensor->Channel[i].UnitStr, ARRAYLEN(pSensor->Channel[i].UnitStr), wchToken, _TRUNCATE);
				}
			}

			// Get User Channel Name
			StrLen = Oak_Retreive_UserString(pSensor->DevicePath, i+1, VOLATILE, Name, 20)+1;
			if (StrLen >1)
				_tcsncpy_s(pSensor->Channel[i].UserChannelName, ARRAYLEN(pSensor->Channel[i].UserChannelName), Name, _TRUNCATE);
			StrLen = Oak_Retreive_UserString(pSensor->DevicePath, i+1, NONVOLATILE, Name, 20)+1;
			if (StrLen >1)
				_tcsncpy_s(pSensor->Channel[i].UserChannelName_NV, ARRAYLEN(pSensor->Channel[i].UserChannelName_NV), Name, _TRUNCATE);
		}
	}


	CloseHandle(DeviceHandle);
}

//-------------------------------------------------------------------------------------------------------------------------
extern "C" __declspec(dllexport) void Oak_Initialize(void)

{
	__DEBUG(); // register notification function, search for all Oak Sensors
}

//-------------------------------------------------------------------------------------------------------------------------
extern "C" __declspec(dllexport) bool Oak_Feature (LPCTSTR DevicePath, BYTE RptBuf[33], bool ExpectResult)
// Send / Receive a Feature Report to/from the sensor device
{
		GUID	HidGuid;
		HANDLE	DeviceHandle;
		bool	Success;
		BYTE	DummyBuf[33] = {0x00, 0x00};
		DWORD	Timeout;
		DWORD	Err;

		// ---- Open Device Handle and File
		HidD_GetHidGuid(&HidGuid);
		DeviceHandle = CreateFile ( DevicePath,
									GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
									(LPSECURITY_ATTRIBUTES)NULL, OPEN_EXISTING, 0, NULL);
		Err =GetLastError();
		if (DeviceHandle == INVALID_HANDLE_VALUE)
		{
			ATLTRACE("Cannot open USB Device\n");
			return FALSE;
		}
		
		// ---- Wait, if Feature request is still pending
		Timeout = GetTickCount()+5000;
		do
		{
			Success = HidD_GetFeature (DeviceHandle, DummyBuf, 33) && (DummyBuf[1] == 0xFF);
			if (!Success)
				Sleep(10);
		} while ( !Success && (GetTickCount() < Timeout) );
		if (!Success)
		{
			ATLTRACE("USB Device did not finish previous Feature command\n");
			CloseHandle(DeviceHandle);
			return FALSE; 
		}
		// ---- send command
		Success = HidD_SetFeature (DeviceHandle, RptBuf, 33) != 0;
		if (!Success)
		{
			ATLTRACE("Cannot Send Feature Report\n");
			CloseHandle(DeviceHandle);
			return FALSE; 
		}

		// ---- wait for the device's answer if applicable
		//      This is the case when data byte 0 of the received report is 0xff
		if (ExpectResult)
		{ 
			Timeout = GetTickCount()+5000;
			do
			{
				Success = HidD_GetFeature (DeviceHandle, RptBuf, 33) && (RptBuf[1] == 0xFF);
				if (!Success)
					Sleep(10);
			} while ( !Success && (GetTickCount() < Timeout) );
			if (!Success)
			{
				ATLTRACE("Cannot Receive Feature Report\n");
				CloseHandle(DeviceHandle);
				return FALSE;
			}
		}

		CloseHandle(DeviceHandle);
		return TRUE;
}

//-------------------------------------------------------------------------------------------------------------------------------
extern "C" __declspec(dllexport) void Oak_Register_Callback  ( tNotifyFct OnConnect, tNotifyFct OnDisconnect,
							  WORD PID, WORD REV, LPTSTR SN, 
							  LPTSTR DeviceName, LPTSTR ChannelName, 
							  LPTSTR UserDeviceName, LPTSTR UserChannelName,
							  LPARAM UserParam)
// Register a Notify function that will be called, when a device is connected that matches all the filter parameters
// to ignore a filter parameter, set it to 0 or NULL respectively
{
	tOakFilter	Filter;
	
	memset( &Filter, 0x00, sizeof(tOakFilter) );

	Filter.PID = PID;
	Filter.REV = REV;

	if (SN != NULL)
		_tcsncpy_s(Filter.SN, ARRAYLEN(Filter.SN), SN, _TRUNCATE);
	if (DeviceName != NULL)
		_tcsncpy_s(Filter.DeviceName, ARRAYLEN(Filter.DeviceName), DeviceName, _TRUNCATE);
	if (ChannelName != NULL)
		_tcsncpy_s(Filter.ChannelName, ARRAYLEN(Filter.ChannelName), ChannelName, _TRUNCATE);
	if (UserDeviceName != NULL)
		_tcsncpy_s(Filter.UserDeviceName, ARRAYLEN(Filter.UserChannelName), UserDeviceName, _TRUNCATE);
	if (UserChannelName != NULL)
		_tcsncpy_s(Filter.UserChannelName, ARRAYLEN(Filter.UserChannelName), UserChannelName, _TRUNCATE);

	Filter.OnConnect    = OnConnect;
	Filter.OnDisconnect = OnDisconnect;
	Filter.UserParam    = UserParam;

	Filter_OnDevChange.push_back(Filter);
}

//-------------------------------------------------------------------------------------------------------------------------
extern "C" __declspec(dllexport) void Oak_DestroyNotifications(void)
// Clear all callback functions that are registered to be called on Connect/Disconnect of an Oak Sensor device
{
	Filter_OnDevChange.clear();
}

//-------------------------------------------------------------------------------------------------------------------------
static bool Oak_Match (tOakSensor *Sensor, tOakFilter *Filter)
{
	bool Match;

	// Verify if the sensor matches a registered notification filter
	Match = TRUE;
	Match &= (Filter->PID==0)					|| Filter->PID == Sensor->PID;
	Match &= (Filter->REV==0)					|| Filter->REV == Sensor->REV;
	Match &= (Filter->SN[0]==0)					|| (_tcscmp(Filter->SN, Sensor->SN)==0);
	Match &= (Filter->DevicePath[0]==0)			|| (_tcsicmp(Filter->DevicePath, Sensor->DevicePath)==0);
	Match &= (Filter->DeviceName[0]==0)			|| (_tcscmp(Filter->DeviceName, Sensor->DeviceName)==0);
	Match &= (Filter->UserDeviceName[0]==0)		|| (_tcscmp(Filter->UserDeviceName, Sensor->UserDeviceName)==0);
	for (int ch=0; ch<Sensor->NumChannels; ch++)
	{
		Match &= (Filter->ChannelName[0]==0)	|| (_tcscmp(Filter->ChannelName, Sensor->Channel[ch].ChannelName)==0);
		Match &= (Filter->UserChannelName[0]==0)|| (_tcscmp(Filter->UserChannelName, Sensor->Channel[ch].UserChannelName)==0);
	}
	return Match;
}
//-------------------------------------------------------------------------------------------------------------------------
extern "C" __declspec(dllexport) void Oak_NotifyConnectedDevices(void)
// Call notify function for all currently connected devices that match the registered filter criteria
{
	bool								bResult;
	USHORT								usiMemberIndex;
	HANDLE								DeviceInfoSet;
	GUID								HidGuid;
	SP_DEVICE_INTERFACE_DATA			DeviceInterfaceData;
	PSP_DEVICE_INTERFACE_DETAIL_DATA	DetailData;
	tOakSensor							Sensor;
	std::vector<tOakFilter>::iterator	itFilter;

	HidD_GetHidGuid(&HidGuid);				// Get the GUID for all system HIDs.

	// Returns a handle to a device information set for all installed devices.
	DeviceInfoSet = SetupDiGetClassDevs(&HidGuid, NULL, NULL, DIGCF_PRESENT|DIGCF_DEVICEINTERFACE);

	DeviceInterfaceData.cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);
	usiMemberIndex = 0;
		// On return, DeviceInterfaceData contains the handle to a 
		// SP_DEVICE_INTERFACE_DATA structure for a detected device.
	bResult = SetupDiEnumDeviceInterfaces(DeviceInfoSet, 0, &HidGuid, usiMemberIndex, &DeviceInterfaceData) != 0;
	
	// Loop once for every HID device interface.
	while(bResult != 0)	//A device has been detected, so get more information about it.
	{
		ULONG Length;
		ULONG Required;
		
		memset( &Sensor, 0x00, sizeof(tOakSensor) ); // clear all fields of the Sensor variable
	
		// Returns an SP_DEVICE_INTERFACE_DETAIL_DATA structure containing information about a device.
		// To retrieve the information, call this function twice. The first time returns the size of 
		// the structure in Length. The second time returns a pointer to the data in DeviceInfoSet.
		bResult = SetupDiGetDeviceInterfaceDetail(DeviceInfoSet, &DeviceInterfaceData, NULL, 0, &Length, NULL) != 0;

		Required = Length;

		DetailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA)malloc(Length);	// Allocate memory for the hDevInfo structure, using the returned Length.
		DetailData->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);	// Set cbSize in the DetailData structure.

		// Call the function again, this time passing it the returned buffer size.
		bResult = SetupDiGetDeviceInterfaceDetail(DeviceInfoSet, &DeviceInterfaceData, DetailData, Required, &Length, NULL) != 0;

		// Retreive Sensor information
		Oak_Retreive_SensorInfo( DetailData->DevicePath, &Sensor );
		if (Sensor.VID == VENDOR_ID_TORADEX)
		{
			// Retreive additional information for Toradex Oak Sensor devices.
			Oak_Retreive_SpecificSensorInfo( &Sensor );

			// Verify if the newly found sensor matches one or more registered notification filters
			for (itFilter=Filter_OnDevChange.begin(); itFilter != Filter_OnDevChange.end(); itFilter++)
				// call the registered notification function, if filter matches
				if ( Oak_Match(&Sensor, &(*itFilter)) )
				{
					if ( SensorList_FindPath(&Sensor.DevicePath[0]) == NULL )
						SensorList_Add(&Sensor);
					itFilter->OnConnect(&Sensor, itFilter->UserParam);
				}
		}

		Oak_DestroySensor(&Sensor);

		// Free memory that was allocated to hold the sensor interface's data
		free(DetailData);
		
		// On return, DeviceInterfaceData contains the handle to a 
		// SP_DEVICE_INTERFACE_DATA structure for a detected device.
		bResult = SetupDiEnumDeviceInterfaces(DeviceInfoSet, 0, &HidGuid, usiMemberIndex, &DeviceInterfaceData) != 0;
		usiMemberIndex++;
	}

	SetupDiDestroyDeviceInfoList(DeviceInfoSet);
}

//-------------------------------------------------------------------------------------------------------------------------
LRESULT CALLBACK Oak_Hotplug_WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
// Create a window and its Message loop to handle Device arrival / removal notifications
{
	PDEV_BROADCAST_HDR					lpdb = 0;
	PDEV_BROADCAST_DEVICEINTERFACE		lpdbi = 0;
	tOakSensor							Sensor = { 0 };
	std::vector<tOakFilter>::iterator	itFilter;

	switch (message)
	{
		case WM_DEVICECHANGE:
			// ---- if message of interest, retreive the device's pathname
			switch (wParam)
			{
				case DBT_DEVICEARRIVAL:
				case DBT_DEVICEREMOVECOMPLETE:
					lpdb  = (PDEV_BROADCAST_HDR)lParam;
					if (lpdb->dbch_devicetype == DBT_DEVTYP_DEVICEINTERFACE)
						lpdbi = (PDEV_BROADCAST_DEVICEINTERFACE)lParam;
			}
			
			// ---- now handle device arrival differently than device removal
			if (lpdbi != NULL)
			{
				switch (wParam)
				{
					case DBT_DEVICEARRIVAL:
						// Retreive Sensor information
						memset( &Sensor, 0x00, sizeof(tOakSensor) );
						Oak_Retreive_SensorInfo(lpdbi->dbcc_name, &Sensor);
						if (Sensor.VID == VENDOR_ID_TORADEX)
						{
							// Retreive additional information for Toradex Oak Sensor devices.
							Oak_Retreive_SpecificSensorInfo( &Sensor );

							// Verify if the newly found sensor matches one or more registered notification filters
							for (itFilter=Filter_OnDevChange.begin(); itFilter != Filter_OnDevChange.end(); itFilter++)
								// call the registered notification function, if filter matches
								if ( Oak_Match(&Sensor, &(*itFilter)) )
								{
									if ( SensorList_FindPath(&Sensor.DevicePath[0]) == NULL )
										SensorList_Add(&Sensor);
									
									itFilter->OnConnect(&Sensor, itFilter->UserParam);
								}
						}

						Oak_DestroySensor(&Sensor);
						break;

					case DBT_DEVICEREMOVECOMPLETE:
						tOakSensor *RemovedSensor = SensorList_FindPath(lpdbi->dbcc_name);
						if (RemovedSensor != NULL)
						{
							//Call Notification Functions for disconnect.
							for (itFilter=Filter_OnDevChange.begin(); itFilter != Filter_OnDevChange.end(); itFilter++)
								// call the registered notification function, if filter matches
								if ( Oak_Match(RemovedSensor, &(*itFilter)) )
								{
									itFilter->OnDisconnect(RemovedSensor, itFilter->UserParam);
								}
						
							// remove sensor from the list of sensors that are registered for disconnect events
							SensorList_RemoveSensor(RemovedSensor->DevicePath);
						}
						break;
				}
				break;
			}
		default:
			  return DefWindowProc(hWnd, message, wParam, lParam);
   }
   return 0;
}
//-------------------------------------------------------------------------------------------------------------------------
DWORD WINAPI Oak_Hotplug_Setup(LPVOID lpParameter)
{
	HINSTANCE						Instance = (HINSTANCE)lpParameter;
	ATOM							Atom;
	WNDCLASSEX						WndClass;
	HWND							hWnd;
	GUID							HidGuid;
	HDEVNOTIFY						DeviceNotificationHandle;
	DEV_BROADCAST_DEVICEINTERFACE	DevBroadcastDeviceInterface;
	BOOL							bRet;
	MSG								msg;

	// ---- Register new Window class
	WndClass.cbSize			= sizeof(WndClass);
    WndClass.style			= 0;
    WndClass.lpfnWndProc	= Oak_Hotplug_WndProc;
    WndClass.cbClsExtra		= 0;
	WndClass.cbWndExtra		= 0;
    WndClass.hInstance		= Instance;
    WndClass.hIcon			= NULL;
    WndClass.hCursor		= NULL;
    WndClass.hbrBackground	= (HBRUSH)(COLOR_BACKGROUND+1);
    WndClass.lpszMenuName	= NULL;
    WndClass.lpszClassName	= _T("OakHotplug");
    WndClass.hIconSm		= NULL;

	Atom = RegisterClassEx(&WndClass);

	// ---- create window
	hWnd = CreateWindowEx(WS_EX_TRANSPARENT | WS_EX_TOOLWINDOW, _T("OakHotplug"), NULL, 
						WS_POPUP, CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, 
						NULL, NULL, Instance, NULL);
	
	// ---- Register to receive device arrival / removal messages
	HidD_GetHidGuid(&HidGuid);
	DevBroadcastDeviceInterface.dbcc_size		= sizeof(DevBroadcastDeviceInterface);
	DevBroadcastDeviceInterface.dbcc_devicetype	= DBT_DEVTYP_DEVICEINTERFACE;
	DevBroadcastDeviceInterface.dbcc_classguid	= HidGuid;
	DeviceNotificationHandle = RegisterDeviceNotification(hWnd, &DevBroadcastDeviceInterface, DEVICE_NOTIFY_WINDOW_HANDLE);

    // ---- Main message loop:
	while(bRet = GetMessage( &msg, NULL, 0, 0 ) != 0)
	{ 
		if (bRet == -1)
		{
			// handle the error and possibly exit
		}
		else
		{
			TranslateMessage(&msg); 
			DispatchMessage(&msg); 
		}
	}
	return 0; // Thread Successful
}
//-------------------------------------------------------------------------------------------------------------------------
#if _MSC_VER >= 1300    // for VC 7.0
  // from ATL 7.0 sources
  #ifndef _delayimp_h
  extern "C" IMAGE_DOS_HEADER __ImageBase;
  #endif
#endif

HMODULE GetCurrentModule()
{
#if _MSC_VER < 1300    // earlier than .NET compiler (VC 6.0)

  // Here's a trick that will get you the handle of the module
  // you're running in without any a-priori knowledge:
  // http://www.dotnet247.com/247reference/msgs/13/65259.aspx

  MEMORY_BASIC_INFORMATION mbi;
  static int dummy;
  VirtualQuery( &dummy, &mbi, sizeof(mbi) );

  return reinterpret_cast<HMODULE>(mbi.AllocationBase);

#else    // VC 7.0

  // from ATL 7.0 sources

  return reinterpret_cast<HMODULE>(&__ImageBase);
#endif
}
//-------------------------------------------------------------------------------------------------------------------------
extern "C" __declspec(dllexport) void Oak_Hotplug_Enable(void)
{
	HANDLE Oak_Hotplug_Thread;
	DWORD IDThread;
	
	//HANDLE h1 = GetModuleHandle(NULL);
	//HANDLE h2 = AfxGetInstanceHandle();
	HANDLE h3 = GetCurrentModule();

	Oak_Hotplug_Thread = CreateThread(NULL, 0, &Oak_Hotplug_Setup, h3, 0, &IDThread);
}
//===========================================================================================
// DEBUG functions
//===========================================================================================
void  __stdcall DeviceConnect(PtOakSensor Sensor, LPARAM UserParam)
// This function will be registered ( Oak_Register_ConnectFct() ) to be called, when a new sensor is attached
{
//	BYTE buffer[33] = {0, 0x08,0x00,0x01}; // Command to start "smile" function :-)
//	Oak_Feature(Sensor.DevicePath, buffer, FALSE);
	ATLTRACE("Debug.OnConnect: VID:%4X, PID:%4X, REV:%4X, %s [0x%08X]\n", Sensor->VID, Sensor->PID, Sensor->REV, Sensor->DeviceName, UserParam);
}
//-------------------------------------------------------------------------------------------------------------------------
void  __stdcall DeviceDisconnect(PtOakSensor Sensor, LPARAM UserParam)
// This function will be registered ( Oak_Register_ConnectFct() ) to be called, when a new sensor is attached
{
	ATLTRACE("Debug.OnDisconnect: VID:%4X, PID:%4X, REV:%4X, %s [0x%08X]\n", Sensor->VID, Sensor->PID, Sensor->REV, Sensor->DeviceName, UserParam);
}

//-------------------------------------------------------------------------------------------------------------------------
void __DEBUG()
// Only for debug purposes: register a few callback functions,
// then trigger the handler to call the registered connect functions for all devices that are already attached
{

//	tNotifyFct fct = &NotifyFunction;
	Oak_Register_Callback( &DeviceConnect, &DeviceDisconnect,0x0002,0x0000,NULL,NULL,NULL,NULL,NULL, 0x12345678); 
	Oak_Register_Callback( &DeviceConnect, &DeviceDisconnect,0x0000,0x0000,NULL,_T("Pressure Sensor"),NULL,NULL,NULL, 0);
	Oak_NotifyConnectedDevices();
}

//===========================================================================================
// list of Sensors
//===========================================================================================

typedef struct sElement
{
	tOakSensor		 Sensor;
	struct sElement *next;
} tElement;

tElement *Root = NULL;

//-------------------------------------------------------------------------------------------------------------------------
void SensorList_Add (tOakSensor *NewSensor)
{
	tElement *Current;

	// --- allocate memory for a new element. Let "Current" point to it.
	if (Root == NULL)
	{	// if list is empty
		Root = (tElement*)malloc(sizeof(tElement));
		Current = Root;
	}
	else
	{	// if list is not empty
		Current = Root;
		while (Current->next != NULL)
			Current = Current->next;
		Current->next = (tElement*)malloc(sizeof(tElement));
		Current = Current->next;
	}
	
	// --- fill new element with data
	Current->next = NULL;
	memcpy_s( &Current->Sensor, sizeof(Current->Sensor), NewSensor, sizeof(*NewSensor) );
}

//-------------------------------------------------------------------------------------------------------------------------
// TODO: call this function at an appropriate place!

//-------------------------------------------------------------------------------------------------------------------------
void SensorList_Destroy()
{
	tElement *Current, *Next;

	Current = Root;
	Root = NULL;

	while (Current != NULL)
	{
		Next = Current->next;
		free (Current);
		Current = Next;
	}
}

//-------------------------------------------------------------------------------------------------------------------------
tOakSensor *SensorList_FindPath(TCHAR *ComparePath)
{
	bool		 Found = FALSE;
	tElement	*Current;

	Current = Root;
	while ( !Found && (Current != NULL) )
	{
		Found |= (_tcsicmp(Current->Sensor.DevicePath, ComparePath)==0);
		if (!Found)
			Current = Current->next;
	}

	return &Current->Sensor;  // returns pointer to the demanded Sensor, NULL if not found.
}

//-------------------------------------------------------------------------------------------------------------------------
void SensorList_RemoveSensor(TCHAR *RemovePath)
{
	bool		 Found = FALSE;
	tElement	*Current, *Previous;

	Previous = NULL;
	Current = Root;
	while ( !Found && (Current != NULL) )
	{
		Found |= (_tcsicmp(Current->Sensor.DevicePath, RemovePath)==0);
		if (!Found)
		{
			Previous = Current;		
			Current = Current->next;
		}
	}

	if (Found)
	{
		if (Previous == NULL)
			Root = Current->next;
		else
			Previous->next = Current->next;

		free(Current);
	}
}


//---------------------------------------------------------------------------
// Blocking functions that return directly with a result
//---------------------------------------------------------------------------
extern "C" __declspec(dllexport) bool Oak_FindSensor(
									  WORD PID, WORD REV, LPCTSTR SN,					// filters to be set
									  LPCTSTR DeviceName, LPCTSTR ChannelName,			// 0 or NULL acts as wildcard.
									  LPCTSTR UserDeviceName, LPCTSTR UserChannelName,
									  PtOakSensor SensorFound)							// this is the return value
// Search for a sensor and result a structure containing information about it.
// if the provided filters fit for more than one sensor, only one (random) sensor is returned.
// The function returns true, if a sensor matched the provided filters.
{
	bool								bResult;
	USHORT								usiMemberIndex;
	HANDLE								DeviceInfoSet;
	GUID								HidGuid;
	SP_DEVICE_INTERFACE_DATA			DeviceInterfaceData;
	PSP_DEVICE_INTERFACE_DETAIL_DATA	DetailData;
	bool								found = false;

	tOakFilter	Filter;
	
	// --- setup the filter with the provided function parameters
	memset( &Filter, 0x00, sizeof(tOakFilter) );
	Filter.PID = PID;
	Filter.REV = REV;
	if (SN != NULL)
		_tcsncpy_s(Filter.SN, ARRAYLEN(Filter.SN), SN, _TRUNCATE);
	if (DeviceName != NULL)
		_tcsncpy_s(Filter.DeviceName, ARRAYLEN(Filter.DeviceName), DeviceName, _TRUNCATE);
	if (ChannelName != NULL)
		_tcsncpy_s(Filter.ChannelName, ARRAYLEN(Filter.ChannelName), ChannelName, _TRUNCATE);
	if (UserDeviceName != NULL)
		_tcsncpy_s(Filter.UserDeviceName, ARRAYLEN(Filter.UserChannelName), UserDeviceName, _TRUNCATE);
	if (UserChannelName != NULL)
		_tcsncpy_s(Filter.UserChannelName, ARRAYLEN(Filter.UserChannelName), UserChannelName, _TRUNCATE);

	// --- prepare getting data about each connected HID device
	memset( SensorFound, 0x00, sizeof(SensorFound) ); // clear all fields of the Sensor variable
	HidD_GetHidGuid(&HidGuid);				// Get the GUID for all system HIDs.

	// Returns a handle to a device information set for all installed devices.
	DeviceInfoSet = SetupDiGetClassDevs(&HidGuid, NULL, NULL, DIGCF_PRESENT|DIGCF_DEVICEINTERFACE);

	DeviceInterfaceData.cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);
	usiMemberIndex = 0;
		// On return, DeviceInterfaceData contains the handle to a 
		// SP_DEVICE_INTERFACE_DATA structure for a detected device.
	bResult = SetupDiEnumDeviceInterfaces(DeviceInfoSet, 0, &HidGuid, usiMemberIndex, &DeviceInterfaceData) != 0;
	
	// --- Loop once for every HID device interface.
	while(bResult && !found)	//A device has been detected, so get more information about it.
	{
		ULONG Length;
		ULONG Required;
		
		// Returns an SP_DEVICE_INTERFACE_DETAIL_DATA structure containing information about a device.
		// To retrieve the information, call this function twice. The first time returns the size of 
		// the structure in Length. The second time returns a pointer to the data in DeviceInfoSet.
		bResult = SetupDiGetDeviceInterfaceDetail(DeviceInfoSet, &DeviceInterfaceData, NULL, 0, &Length, NULL) != 0;

		Required = Length;

		DetailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA)malloc(Length);	// Allocate memory for the hDevInfo structure, using the returned Length.
		DetailData->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);	// Set cbSize in the DetailData structure.

		// Call the function again, this time passing it the returned buffer size.
		bResult = SetupDiGetDeviceInterfaceDetail(DeviceInfoSet, &DeviceInterfaceData, DetailData, Required, &Length, NULL) != 0;

		// Retreive Sensor information
		Oak_Retreive_SensorInfo( DetailData->DevicePath, SensorFound );
		if (SensorFound->VID == VENDOR_ID_TORADEX)
		{
			// Retreive additional information for Toradex Oak Sensor devices.
			Oak_Retreive_SpecificSensorInfo( SensorFound );

			// Verify if the newly found sensor matches the provided filter
			if ( Oak_Match(SensorFound, &Filter) )
				found = true;
		}

		// Free memory that was allocated to hold the sensor interface's data
		free(DetailData);
		
		// On return, DeviceInterfaceData contains the handle to a 
		// SP_DEVICE_INTERFACE_DATA structure for a detected device.
		bResult = SetupDiEnumDeviceInterfaces(DeviceInfoSet, 0, &HidGuid, usiMemberIndex, &DeviceInterfaceData) != 0;
		usiMemberIndex++;
	}

	SetupDiDestroyDeviceInfoList(DeviceInfoSet);

	return found;
}

//-------------------------------------------------------------------------------------------------------------------------
extern "C" __declspec(dllexport) bool Oak_GetCurrentInReport (LPCTSTR DevicePath, BYTE RptBuf[33], BYTE InReportLength)
// Get current device state (last sent report)
// - this function only works for Windows Versions XP and higher
// - do not use this function to continously read data from the sensor - as it always returns the last
//   report received, reports may be lost.
{
	HANDLE	DeviceHandle;
	bool	bSuccess;	
	DWORD	Err;

	DeviceHandle = CreateFile ( DevicePath,
								GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
								(LPSECURITY_ATTRIBUTES)NULL, OPEN_EXISTING, 0, NULL);
	Err =GetLastError();
	if (DeviceHandle == INVALID_HANDLE_VALUE)
	{
		ATLTRACE("Cannot open USB Device\n");
		return false;
	}

	RptBuf[0] = 0;
	bSuccess = HidD_GetInputReport(DeviceHandle, RptBuf, InReportLength) != 0;
	CloseHandle(DeviceHandle);

	return bSuccess;
}

//-------------------------------------------------------------------------------------------------------------------------
extern "C" __declspec(dllexport) bool Oak_GetInReport (LPCTSTR DevicePath, BYTE RptBuf[33], BYTE InReportLength, WORD Timeout_ms)
// Wait for a report to be received and return it.
// The function returns anyway, after "Timeout_ms" has elapsed.
// - do not set the Timeout_ms close to the sensor's report rate. We found that reports may get lost if you do.
{
		HANDLE	DeviceHandle;
		HANDLE  hEventObject;
		OVERLAPPED HIDOverlapped;
		DWORD	uliResult;
		DWORD	Err;
		DWORD	uliNumberOfBytesRead;

		// ---- Open Device Handle and File
		DeviceHandle = CreateFile ( DevicePath,
									GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
									(LPSECURITY_ATTRIBUTES)NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL);
		Err =GetLastError();
		if (DeviceHandle == INVALID_HANDLE_VALUE)
		{
			ATLTRACE("Cannot open USB Device\n");
			return false;
		}

		hEventObject = CreateEvent(NULL, false, true, _T(""));
		HIDOverlapped.hEvent = hEventObject;
		HIDOverlapped.Offset = 0;
		HIDOverlapped.OffsetHigh = 0;
		uliResult = ReadFile(DeviceHandle, RptBuf, InReportLength, &uliNumberOfBytesRead, &HIDOverlapped) != 0;
		uliResult = WaitForSingleObject(hEventObject, Timeout_ms);
		switch (uliResult)
		{
			case WAIT_OBJECT_0 :		// Data read succeeded
            CloseHandle(DeviceHandle);
            CloseHandle(hEventObject);
				return true;
				break;

			case WAIT_TIMEOUT :			// Timeout error
			default :					// Undefined error
			{
				CancelIo(DeviceHandle);
            CloseHandle(DeviceHandle);
            CloseHandle(hEventObject);
				return false;
			}
		}

		return true;
}

//-------------------------------------------------------------------------------------------------------------------------
extern "C" __declspec(dllexport) bool Oak_SendOutReport (LPCTSTR DevicePath, BYTE RptBuf[33], BYTE OutReportLength, WORD Timeout_ms)
// Waits for sending a report and sent it.
// The function returns anyway, after "Timeout_ms" has elapsed.
// - do not set the Timeout_ms close to the sensor's report rate. We found that reports may get lost if you do.
{
		HANDLE	DeviceHandle;
		HANDLE  hEventObject;
		OVERLAPPED HIDOverlapped;
		DWORD	uliResult;
		DWORD	Err;
		DWORD	uliNumberOfBytesWritten;

		// ---- Open Device Handle and File
		DeviceHandle = CreateFile ( DevicePath,
									GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
									(LPSECURITY_ATTRIBUTES)NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL);
		Err =GetLastError();
		if (DeviceHandle == INVALID_HANDLE_VALUE)
		{
			ATLTRACE("Cannot open USB Device\n");
			return false;
		}

		hEventObject = CreateEvent(NULL, false, true, _T(""));
		HIDOverlapped.hEvent = hEventObject;
		HIDOverlapped.Offset = 0;
		HIDOverlapped.OffsetHigh = 0;
		uliResult = WriteFile(DeviceHandle, RptBuf, OutReportLength, &uliNumberOfBytesWritten, &HIDOverlapped) != 0;
		uliResult = WaitForSingleObject(hEventObject, Timeout_ms);
		switch (uliResult)
		{
			case WAIT_OBJECT_0 :		// Data send succeeded
            CloseHandle(DeviceHandle);
            CloseHandle(hEventObject);
				return true;
				break;

			case WAIT_TIMEOUT :			// Timeout error
			default :					// Undefined error
			{
				CancelIo(DeviceHandle);
            CloseHandle(DeviceHandle);
            CloseHandle(hEventObject);
				return false;
			}
		}

		return true;
}

extern "C" __declspec(dllexport) void Oak_Cleanup()
{
   SensorList_Destroy();
}

//-------------------------------------------------------------------------------------------------------------------------
